'// A Mailing List Membership Quicksort Optimization Tool
'// Sorts a list of email addresses.
'// Created 20010311 by Casey Sprangel
'// Written in PowerBasic CC for Windows (http://www.powerbasic.com)
'// Copyright 2001-2008 all rights reserved.
'//--------------------------------------------------------------------------
'// Things to add or change:
'// - resolve all domains found (pre-load DNS cache)
'// - sort ascending or decending
'// - "untidy" option to leave temp data files around after exit
'// - sort based on connect times to domain (was not possible in DOS ver)
'// - sort based on resolve times (was not possible in DOS ver)
'// - check for full disk condition before sort
'// - check for write errors and handle them more gracefully
'// - check for read errors and handle them more gracefully
'// - "clean" all domain names by forcing to lowercase
'//--------------------------------------------------------------------------

#DIM ALL                                              '// force a declare for every variable and function
#TOOLS OFF                                            '// turn off development tools in compiled code

DECLARE FUNCTION exist(f AS STRING) AS LONG
DECLARE FUNCTION chunk(f AS STRING, session AS LONG) AS LONG
DECLARE FUNCTION closed(f AS STRING) AS LONG
DECLARE FUNCTION flip(a AS STRING) AS STRING
DECLARE FUNCTION sortfile(f AS STRING, fi AS STRING, fb AS STRING) AS LONG

%version = 20081031
$title = "ListOpt "
GLOBAL nodupe AS LONG                                 '// flag to show if we filter duplicates

'//---------------------------------------------------------------------------

FUNCTION exist(f AS STRING) AS LONG
  LOCAL i AS LONG
  ON ERROR RESUME NEXT                                '// don't stop for errors (turns off error traps)
  i = GETATTR(f)                                      '// get the file attributes (generate error on failure)
  FUNCTION = (ERRCLEAR = 0)                           '// return true if no errors
END FUNCTION

'//---------------------------------------------------------------------------

FUNCTION closed(f AS STRING) AS LONG
  LOCAL h AS LONG
  LOCAL i AS LONG
  LOCAL buf AS STRING
  IF exist(f) THEN                                    '// does the file exist?
    h = FREEFILE                                      '// get next available file handle
    OPEN f FOR BINARY AS h                            '// open the list file in binary mode
    i = 65536                                         '// start with a 64k chunk size
    IF i > LOF(h) THEN                                '// if file is smaller than chunk size
      i = LOF(h)                                      '// change chunk to total file size
    END IF
    GET$ #h, i, buf                                   '// get a chunk of the file
    CLOSE h                                           '// close the file
    buf = UCASE$(buf)                                 '// uppercase the string for comparison
    i = 0                                             '// start with an empty flag
    IF LEFT$(buf, 8) = "CLOSED" + $CRLF THEN          '// does the file starts with CLOSED?
      i = 1                                           '// the flag is true
    ELSEIF RIGHT$(buf, 8) = $CRLF + "CLOSED" THEN     '// last line of file is CLOSED?
      i = 1                                           '// the flag is true
    ELSE                                              '// everything else, just look for the line by itself
      i = INSTR(UCASE$(buf), $CRLF + "CLOSED" + $CRLF) '// locate the word CLOSED on a line by itself
    END IF
    FUNCTION = i                                      '// pass back what we found
  END IF
END FUNCTION

'//---------------------------------------------------------------------------

FUNCTION sortfile(f AS STRING, fi AS STRING, fb AS STRING) AS LONG  '// quicksorts a file with include and blacklist and returns record count
  REGISTER i AS LONG                                  '// generic pointer (inside CPU register for max speed)
  LOCAL maxlen AS LONG                                '// maximum record length in file
  LOCAL buf AS STRING                                 '// working buffer
  LOCAL buf2 AS STRING                                '// only used for blacklist filtering
  LOCAL blackbuf AS STRING                            '// holds the blacklist records
  LOCAL p AS STRING                                   '// holds plus sign or spacer in lists with suspended members
  LOCAL f2 AS STRING                                  '// temp output file name
  LOCAL h1 AS LONG                                    '// input file handle
  LOCAL h2 AS LONG                                    '// output file handle
  LOCAL h3 AS LONG
  LOCAL rtotal AS LONG                                '// keeps track of the total record count
  LOCAL top AS STRING                                 '// tracks things at the top of the file that don't move
  LOCAL isclosed AS LONG                              '// flag to show that the list file is closed
  LOCAL marker1 AS LONG                               '// flag to show we have reached the end of the CLOSED section
  LOCAL plusfound AS LONG                             '// flag to show suspended accounts are in this file
  LOCAL oldbuf AS STRING                              '// tracks duplicate lines
  LOCAL blackflag AS LONG                             '// black list flag
  LOCAL zapped AS LONG                                '// tracks number of blacklist records removed from master file
  LOCAL dupe AS LONG                                  '// tracks number of duplicate files found
  DIM hold(1) AS STRING                               '// a table for sorting
  LOCAL hold_point AS LONG                            '// counter to remember the highest record in our table
  IF NOT exist(f) THEN                                '// file not here?
    EXIT FUNCTION                                     '// leave - nothing for us to do
  END IF
  buf = ""                                            '// make sure we start with an empty string
  IF LEN(fb) > 0 THEN                                 '// is there a blacklist file?
    i = sortfile(fb, buf, buf)                        '// sort it first (the empty buf strings are place holders)
    IF i > 0 THEN                                     '// are there any records left?
      blackflag = -1                                  '// set flag to true
      h3 = FREEFILE                                   '// get next available file handle
      OPEN fb FOR INPUT AS h3                         '// open the blacklist file (in sorted order)
      IF NOT EOF(h3) THEN                             '// something still in the file?
        LINE INPUT #h3, blackbuf                      '// get a line from the file
        blackbuf = UCASE$(TRIM$(blackbuf))            '// make sure it is uppercase with no leading or trailing spaces
      END IF
    END IF
  END IF
  i = INSTR(-1, f, ".")                               '// locate the last period in the input filename
  IF i > 0 THEN                                       '// is there a period?
    f2 = LEFT$(f, i) + "TMP"                          '// change extension
  ELSE                                                '// no period found
    f2 = f + ".TMP"                                   '// add an extension
  END IF
  IF UCASE$(f2) = UCASE$(f) THEN                      '// are they the same name?
    f2 = LEFT$(f2, LEN(f2) -1)                        '// cut off one letter from the temp filename
  END IF
  isclosed = closed(f)                                '// remember if the list is closed
  IF isclosed THEN                                    '// is this a closed list?
    STDOUT "* Closed list found - protecting moderators."
  END IF
  Marker1 = 0                                         '// set our flag to the starting place
  h1 = FREEFILE                                       '// get next available file handle
  OPEN f FOR INPUT AS h1                              '// open the file in text input mode
  h2 = FREEFILE                                       '// get next available file handle
  OPEN f2 FOR OUTPUT AS h2                            '// open the temp file in text output mode
  DO WHILE NOT EOF(h1)                                '// loop through the input file until we get to the end
    LINE INPUT #h1, buf                               '// get a line of text (CR/LF delim)
    INCR rtotal                                       '// add one to our line counter
    IF UCASE$(LEFT$(buf,12)) = "+ERRORS-TO:=" THEN    '// special setting address for the list
      top = top + buf + $CRLF                         '// add the special line to our file topper buffer
      IF NOT isclosed THEN                            '// the flag is not showing we need special top handling?
        isclosed = -1                                 '// set to true to remember this special record
      END IF
      ITERATE                                         '// back to top of loop
    END IF
    IF isclosed AND NOT Marker1 THEN                  '// are we inside of the CLOSED part?
      top = top + buf + $CRLF                         '// add another line to our topper buffer
      IF UCASE$(buf) = "CLOSED" THEN                  '// did we reach the CLOSE line?
        Marker1 = -1                                  '// set the flag to show we are in normal user territory
      END IF
      ITERATE                                         '// back to top of loop
     END IF
     buf = TRIM$(buf)                                 '// remove any leading or trailing spaces
     IF LEFT$(buf, 1) = "+" THEN                      '// check for disabled accounts (tracks if we need to add spaces later)
       plusfound = -1                                 '// set flag to show suspended accounts are in this file
     END IF
     INCR hold_point                                  '// increment pointer to the next table element
     IF hold_point > UBOUND(hold()) THEN              '// are we pointing past the end of the table?
       REDIM PRESERVE hold(hold_point + 1000)         '// expand the table by 1000 elements
     END IF
     hold(hold_point) = buf                           '// add this element to the table
  LOOP
  IF LEN(fi) > 0 THEN                                 '// is there a file to include?
    CLOSE h1                                          '// close the input file
    OPEN fi FOR INPUT AS h1                           '// open the new input file
    DO WHILE NOT EOF(h1)                              '// loop until we read all of the file
      LINE INPUT #h1, buf                             '// read a line from the file (CR/LF delim)
      INCR rtotal                                     '// add one to our counter
      buf = TRIM$(buf)                                '// remove any leading spaces
      IF LEFT$(buf, 1) = "+" THEN                     '// check for disabled accounts
        plusfound = -1                                '// flag to show suspended accounts are in this file
      END IF
     INCR hold_point                                  '// increment pointer to the next table element
     IF hold_point > UBOUND(hold()) THEN              '// are we pointing past the end of the table?
       REDIM PRESERVE hold(hold_point + 1000)         '// expand the table by 1000 elements
     END IF
     hold(hold_point) = buf                           '// add this element to the table
    LOOP
  END IF
  CLOSE h1                                            '// close input file
  CLOSE h2                                            '// close output file
  IF rtotal > 1 THEN                                  '// is there anything to sort?
    STDOUT "* " + $DQ + f + $DQ + " has " + FORMAT$(rtotal) + " records." '// show user the number of addresses we found
    STDOUT "   Reading";
    IF UBOUND(hold()) > hold_point THEN               '// are there extra "blank" elements in our table?
      REDIM PRESERVE hold(hold_point)                 '// shrink the table to just the elements we need
    END IF
    IF blackflag THEN                                 '// do we have a blacklist open?
      STDOUT ", Filtering";
      IF LEN(blackbuf) > 0 THEN                       '// is there something in our blacklist buffer to check?
        DO                                            '// loop through the blacklist file
          IF RIGHT$(TIME$,1) <> oldbuf THEN           '// has the second changed on the clock?
            CONSOLE NAME $title + ": Filtering [" + FORMAT$(zapped) + " record" + IIF$(zapped<>1,"s","") + " removed]" '// update the title bar
            oldbuf = RIGHT$(TIME$,1)                  '// remember this second
          END IF
          FOR i = 0 TO UBOUND(hold())                 '// loop through our table
            IF hold(i) > "" THEN                      '// is there anything to check in this table entry?
              IF LEFT$(hold(i),1) = "+" THEN          '// is there a plus sign in this record?
                buf = UCASE$(MID$(hold(i),2))         '// forget the plus sign for this compare
              ELSE                                    '// no plus sign?
                buf = UCASE$(hold(i))                 '// use the whole record
              END IF                                  '// is there a plus sign
              IF buf = blackbuf THEN                  '// this entry is on our blacklist
                INCR zapped                           '// count the number of records removed
                hold(i) = ""                          '// blank out this record
'                array delete hold(i)                  '// remove it from the list
'                decr hold_point
              END IF                                  '// is this a blacklisted record?
            END IF                                    '// is there anything in this record?
          NEXT
          IF NOT EOF(h3) THEN                         '// not at the end of the blacklist yet?
            LINE INPUT #h3, blackbuf                  '// get another record from the blacklist file
            blackbuf = UCASE$(TRIM$(blackbuf))        '// make sure we have an easy match
          ELSE                                        '// we have reached the end of our file
            EXIT LOOP                                 '// leave the loop
          END IF
        LOOP                                          '// try to find this record
      END IF                                          '// is buf not empty?
    END IF                                            '// are we looking for blacklist items?
    CONSOLE NAME $title + ": Sorting"
    STDOUT ", Sorting";
    ARRAY SORT hold(), COLLATE UCASE, DESCEND         '// sort the table in case-insensitive descending order
    CONSOLE NAME $title + ": Writing " + f
    STDOUT ", Writing."                               '// tell user what we are doing
    h2 = FREEFILE                                     '// get next available file handle
    OPEN f FOR OUTPUT AS h2                           '// open original file for output
    IF LEN(top) > 0 THEN                              '// this is the part we saved for closed lists
     ?#h2, top;                                       '// the semicolon is because our string already contains line feeds
    END IF
    DO WHILE hold_point > 0                           '// loop until we read all of the table
      buf = hold(hold_point)                          '// get a table element
      DECR hold_point                                 '// move the pointer to the next record
      IF blackbuf <> RIGHT$(TIME$,1) THEN
        CONSOLE NAME $title + ": Writing " + f + " [" + FORMAT$(hold_point) + "]"
        blackbuf = RIGHT$(TIME$,1)
      END IF
      IF LEN(buf) = 0 THEN                            '// is this an empty record?
        ITERATE                                       '// jump back to top of loop and get another
      END IF
      i = INSTR(buf, "@")                             '// find the AT sign
      IF i > 0 THEN                                   '// did we find it?
        IF nodupe THEN                                '// are we checking for dupe names?
          IF UCASE$(buf) = oldbuf THEN                '// does this address equal our previous address?
            INCR dupe                                 '// track number of dupes found
            ITERATE                                   '// go back to top of loop and forget this address
          END IF
        END IF
        IF plusfound THEN                             '// we should pad normal accounts
          IF LEFT$(buf, 1) <> "+" THEN                '// account is not suspended
            p = $SPC                                  '// add leading space
          ELSE                                        '// it is suspended
            p = ""                                    '// add nothing
          END IF
        END IF
      END IF
      IF LEN(buf) > 0 THEN                            '// is there something in the buffer?
        ?#h2, p + TRIM$(buf)                          '// write it to the output file
        oldbuf = UCASE$(buf)                          '// remember this record for the next pass dupe filtering
      END IF
    LOOP
  END IF                                              '// is there anything to sort?
  CLOSE h1                                            '// close input file
  CLOSE h2                                            '// close output file
  IF blackflag THEN                                   '// do we have a blacklist open?
    CLOSE h3                                          '// close this file too
  END IF
  IF exist(f2) THEN                                   '// is backup file found?
   KILL f2                                            '// erase it
  END IF
  IF dupe > 0 THEN
    STDOUT "   Removed " + FORMAT$(dupe) + " duplicate record" + IIF$(dupe <> 1,"s.",".")
  END IF
  IF zapped > 0 THEN
    STDOUT "   Filtered " + FORMAT$(zapped) + " excluded record" + IIF$(zapped <> 1,"s.",".")
  END IF
  FUNCTION = rtotal                                   '// return the total number of records processed
END FUNCTION

'//---------------------------------------------------------------------------

FUNCTION flip(a AS STRING) AS STRING                  '// reverse the username/domain order in an email address
  LOCAL i AS LONG
  i = INSTR(a, $NUL)                                  '// locate null (this is a flipped record already)
  IF i > 0 THEN
    FUNCTION = MID$(a, i + 1) + "@" + LEFT$(a, i - 1) '// return the unflipped string
    EXIT FUNCTION
  END IF
  i = INSTR(a, "@")                                   '// locate at sign
  IF i > 0 THEN
    FUNCTION = LCASE$(MID$(a, i + 1)) + $NUL + LEFT$(a, i - 1) '// return the flipped string (@ becomes null so string does not grow)
    EXIT FUNCTION
  END IF
  FUNCTION = a                                        '// just return what we have
END FUNCTION

'//---------------------------------------------------------------------------

FUNCTION chunk(f AS STRING, session AS LONG) AS LONG  '// changes a file to have a max of "SESSION" messages to the same domain name in a row
  REGISTER i AS LONG                                  '// generic pointer/counter (inside CPU for max speed)
  REGISTER rcount AS LONG                             '// record count
  LOCAL buf AS STRING                                 '// working string buffer
  LOCAL domain AS STRING                              '// holds the domain name for checking
  LOCAL top AS STRING                                 '// holds everything above the "CLOSED" line if there is one
  DIM list(100) AS STRING                             '// holds the list of addresses
  DIM d(10) AS STRING                                 '// holds records with the same domain
  LOCAL dpoint AS LONG                                '// record pointer into the d table
  LOCAL h AS LONG                                     '// file handle
  LOCAL p AS LONG
  IF NOT exist(f) THEN                                '// input file does not exist?
    EXIT FUNCTION                                     '// leave, we can't do any chunking
  END IF
  h = FREEFILE                                        '// get next available file handle
  OPEN f FOR INPUT AS h                               '// open our file for text input
  rcount = 0                                          '// start the record counter at zero
  top = ""                                            '// start with nothing at the top of our list
  DO WHILE NOT EOF(h)                                 '// loop through the file
    LINE INPUT #h, buf                                '// get a record from the file
    IF UCASE$(LEFT$(buf, 9)) = "+WELCOME=" THEN       '// special setting address for the list
      top = top + buf + $CRLF                         '// add the special line to our file topper buffer
      ITERATE                                         '// go back to the top of the loop to get another record
    END IF
    IF UCASE$(LEFT$(buf, 12)) = "+ERRORS-TO:=" THEN   '// special setting address for the list
      top = top + buf + $CRLF                         '// add the special line to our file topper buffer
      ITERATE                                         '// go back to the top of the loop to get another record
    END IF
    IF UCASE$(TRIM$(buf)) = "CLOSED" THEN             '// is this a closed list?
      FOR i = 0 TO rcount                             '// loop through all of the records we have so far
        top = top + flip(list(i)) + $CRLF             '// add them to the top in normal (unflipped) order
        list(i) = ""                                  '// remove this record from our table
      NEXT
      top = top + UCASE$(TRIM$(buf)) + $CRLF          '// add the "CLOSED" word to complete our top
      rcount = 0                                      '// we now have no records in our table
      ITERATE                                         '// go back to the top of the loop to get another record
    END IF
    IF LEN(TRIM$(buf)) > 0 THEN                       '// is there anything on this line?
      list(rcount) = flip(buf)                        '// remember this record in our list
      INCR rcount                                     '// add one to the record counter
      IF rcount > UBOUND(list()) THEN                 '// is counter past the end of our table?
        REDIM PRESERVE list(rcount + 100)             '// make room for 100 more records
      END IF
    END IF                                            '// is record not blank
  LOOP
  CLOSE h                                             '// close the file, we are done for now
  DECR rcount                                         '// remove the last (blank) rcount
  REDIM PRESERVE list(rcount)                         '// make the table only big enough to hold the data with no extra records
  '// we now have all of the records in our table in flipped order
  ARRAY SORT list(), COLLATE UCASE, DESCEND           '// sort the array into case insensitive domain name order
  dpoint = 0                                          '// start the pointer at the first record
  d(0) = list(0) + CHR$(1)                            '// the first domain is known
  list(0) = ""                                        '// we don't need this record anymore
  FOR i = 1 TO rcount                                 '// loop through all of the records
    IF UCASE$(LEFT$(list(i), INSTR(list(i) + $NUL, $NUL))) = _
       UCASE$(LEFT$(d(dpoint), INSTR(d(dpoint) + $NUL, $NUL))) THEN '// we have a match
      d(dpoint) = d(dpoint) + list(i) + CHR$(1)       '// add this record to our domain list will ASCII 1 as delim
      list(i) = ""                                    '// remove that record from the list
    END IF
    IF list(i) > "" THEN                              '// the domain is not in our list yet?
      INCR dpoint                                     '// move pointer up
      REDIM PRESERVE d(dpoint)                        '// expand the table
      d(dpoint) = list(i) + CHR$(1)                   '// add this domain to our list of known domains
      list(i) = ""                                    '// remove it from the main list
    END IF
  NEXT
  ERASE list()                                        '// remove this table from memory
  '// the list should be in domain grouped order by now
  IF dpoint = 0 THEN                                  '// all records are in the same domain?
    EXIT FUNCTION                                     '// we can't do anything with this list so leave things unchanged
  END IF
  FUNCTION = dpoint                                   '// return the number of unique domains we found
  IF rcount < 201 THEN                                '// is the list too short for grouping?
    STDOUT "+ Error: Not enough names to group."
    EXIT FUNCTION
  END IF
  FOR i = 0 TO dpoint                                 '// loop through the list of domains
    h = TALLY(d(i), CHR$(1))                          '// find the number of records in this domain
    d(i) = FORMAT$(h,"00000000") + d(i)               '// add the record count
  NEXT
  ARRAY SORT d(), DESCEND
  FOR i = 0 TO dpoint
    d(i) = MID$(d(i),9)                               '// trim off the numbers
  NEXT
  FOR i = 0 TO dpoint                                 '// loop through list of domains we know
    IF TALLY(d(i), CHR$(1)) > session THEN            '// more than maximum session size?
      h = INSTR(d(i), CHR$(1))                        '// locate first delim
      DO WHILE TALLY(LEFT$(d(i), h), CHR$(1)) < session '// loop until we have "SESSION" delims
        h = INSTR(h + 1,d(i), CHR$(1))                '// find next delim
      LOOP
      INCR dpoint                                     '// move pointer up
      REDIM PRESERVE d(dpoint)                        '// make room for new domain group
      IF i + 10 > dpoint THEN                         '// is skipping a record going to put us past the end of the table?
        d(dpoint) = MID$(d(i), h + 1)                 '// move the extra records here
      ELSE                                            '// there is room to insert a record
        ARRAY INSERT d(i + 10), MID$(d(i), h + 1)     '// move the extra records here
      END IF
      d(i) = LEFT$(d(i), h)                           '// keep just the records that fit our session size
    END IF                                            '// more than max session size
  NEXT
  h = FREEFILE                                        '// get next available file handle
  OPEN f FOR OUTPUT AS h                              '// open the file for text output
  ?#h, top;                                           '// write the top of our file (if there is one)
  FOR rcount = 0 TO dpoint                            '// walk through our domains
    DO WHILE LEN(d(rcount)) > 0
      i = INSTR(d(rcount) + CHR$(1), CHR$(1))         '// find delim offset or end of string
      ?#h, flip(LEFT$(d(rcount), i - 1))              '// write the record in normal format
      d(rcount) = MID$(d(rcount), i + 1)              '// trim off the part we just wrote to the file
    LOOP
  NEXT
  CLOSE h
END FUNCTION

'//---------------------------------------------------------------------------

FUNCTION PBMAIN() AS LONG                             '// program starts here
  DIM c AS STRING                                     '// holds the command line
  DIM file_main AS STRING                             '// holds the main file name
  DIM file_include AS STRING                          '// holds the include file name
  DIM i AS LONG                                       '// generic pointer
  DIM blackfile AS STRING                             '// holds the blacklist filename
  DIM stime AS LONG                                   '// keeps track of the time we do work
  DIM session AS LONG                                 '// flag or count to show if we chunk the list by domain groups
  nodupe = 0

  STDOUT "Casey's IPAD Mailing List Optimizer version " + FORMAT$(%version)
  c = COMMAND$                                      '// get command line arguments
  IF c = "" OR INSTR(c, "?") > 0 THEN               '// did user ask for help?
    STDOUT "Copyright 2001-2008. All rights reserved."
    STDOUT ""
    STDOUT "This tool sorts a members list file of email addresses with optional"
    STDOUT "include and exclude lists."
    STDOUT ""
    STDOUT "Syntax:"
    STDOUT "  LISTOPT <listname.lst> [<include.lst>] [/B=<exclude.lst>] [/ND] [/G[#]]"
    STDOUT "   /B=<file> = Blacklist file."
    STDOUT "   /ND       = Remove duplicate entries."
    STDOUT "   /G[#]     = Sort domains into groups. Optional number sets session size."
    STDOUT "               Default session size is 10."
    EXIT FUNCTION                                   '// leave the program
  END IF
  i = INSTR(UCASE$(c),"/B=")                        '// locate switch for "don't filter duplicate names"
  IF i > 0 THEN                                     '// did we find the switch?
    blackfile = TRIM$(MID$(c, i + 3))               '// get the file name part
    IF LEFT$(blackfile,1) = $DQ THEN                '// does filename start with double quote?
      blackfile = LEFT$(blackfile, INSTR(2, blackfile, $DQ)) '// only keep up to the next double quote
    ELSE
      IF INSTR(blackfile, $SPC) > 0 THEN            '// is there a space in this filename?
        blackfile = LEFT$(blackfile, INSTR(blackfile, $SPC) - 1) '// cut off up to the next space
      END IF
    END IF
    c = TRIM$(LEFT$(c, i-1) + MID$(c, i + 3 + LEN(blackfile))) '// cut our parm out from the commannd line
    blackfile = TRIM$(blackfile, $DQ)               '// remove the double quotes, if any
    IF NOT exist(blackfile) THEN                    '// file not found?
      STDOUT "* Error: exclude file (" + blackfile + ") not found.
      FUNCTION = 1                                  '// set errorlevel 1
      EXIT FUNCTION                                 '// exit the program
    END IF
  END IF
  i = INSTR(UCASE$(c), "/ND")                       '// locate switch for "don't filter duplicate names"
  IF i > 0 THEN                                     '// is command line switch found?
    nodupe = -1                                     '// set flag to true
    c = TRIM$(LEFT$(c, i - 1) + MID$(c, i + 3))     '// remove swtich from command line
  END IF
  i = INSTR(UCASE$(c), "/G")                        '// locate switch for "session chunk"
  IF i > 0 THEN                                     '// is command line switch found?
    IF VAL(MID$(c,i+2)) > 0 THEN
      session = INT(VAL(MID$(c,i + 2)))             '// get the value of the session
      c = TRIM$(LEFT$(c, i - 1) + MID$(c, INSTR(i + 2, c + $SPC, $SPC))) '// remove swtich from command line
    ELSE
      session = -1                                  '// set flag to true
      c = TRIM$(LEFT$(c, i - 1) + MID$(c, i + 2))   '// remove swtich from command line
    END IF
  END IF
  i = INSTR(c, $SPC)                                '// is there a space in our file name?
  IF i > 0 THEN                                     '// space found in filename?
    file_include = LTRIM$(MID$(c, i))               '// get second file name
    c = RTRIM$(LEFT$(c, i))                         '// keep just the first file name
    IF INSTR(file_include, $SPC) > 0 THEN           '// still spaces in filename? (needs check for quoted filename here)
      STDOUT "* Error: Include filename may not contain spaces."
      STDOUT "  Use /? on the command line for help."
      FUNCTION = 1                                  '// exit with ERRORLEVEL 1 set
      EXIT FUNCTION                                 '// leave the program
    END IF
  END IF
  file_main = c                                     '// whatever is left is our main file name
  IF NOT exist(file_main) THEN                      '// first file not found?
    STDOUT "* Error: File not found " + $DQ + file_main + $DQ
    FUNCTION = 1                                    '// exit with ERRORLEVEL 1 set
    EXIT FUNCTION                                   '// leave the program
  END IF
  IF LEN(file_include) > 0 THEN                     '// is file name 2 not empty?
    IF NOT exist(file_include) THEN                 '// is file name 2 here?
      STDOUT "*Error: File not found " + $DQ + file_include + $DQ
      FUNCTION = 1                                  '// exit with ERRORLEVEL 1 set
      EXIT FUNCTION                                 '// leave the program
    END IF
  END IF
  CONSOLE NAME $title + ": Sorting " + file_main
  stime = TIMER                                     '// remember the start time
  sortfile file_main, file_include, blackfile       '// sort the primary file file_main after including file_include and filtering BLACKFILE
  IF session < 0 THEN                               '// did user ask for grouping without providing a number?
    session = 10                                    '// the default session size is 10
  END IF
  IF session > 0 THEN                               '// do we chunk this list by domain name groups?
    CONSOLE NAME $title + ": Grouping"
    i = chunk(file_main, session)                   '// chunk the file based on session size
    STDOUT "*  Found " + FORMAT$(i, "0,") + " different domain name" + IIF$(i = 1,"","s") + "."
  END IF
  STDOUT "Runtime: " + FORMAT$(TIMER - stime, "0.0") + " seconds."
END FUNCTION                                        '// leave program
